@@ -1,7 +1,9 @@
 # $Id: ethernet.py 65 2010-03-26 02:53:51Z dugsong $
 # -*- coding: utf-8 -*-
-"""Ethernet II, LLC (802.3+802.2), LLC/SNAP, and Novell raw 802.3,
-with automatic 802.1q, MPLS, PPPoE, and Cisco ISL decapsulation."""
+"""
+Ethernet II, LLC (802.3+802.2), LLC/SNAP, and Novell raw 802.3,
+with automatic 802.1q, MPLS, PPPoE, and Cisco ISL decapsulation.
+"""
 from __future__ import print_function
 from __future__ import absolute_import
 
@@ -10,15 +12,8 @@ from zlib import crc32
 
 from . import dpkt
 from . import llc
-from .compat import compat_ord, iteritems
-
-try:
-    isinstance("", basestring)
-    def isstr(s):
-        return isinstance(s, basestring)
-except NameError:
-    def isstr(s):
-        return isinstance(s, str)
+from .utils import mac_to_str
+from .compat import compat_ord, iteritems, isstr
 
 ETH_CRC_LEN = 4
 ETH_HDR_LEN = 14
@@ -30,6 +25,7 @@ ETH_MTU = (ETH_LEN_MAX - ETH_HDR_LEN - ETH_CRC_LEN)
 ETH_MIN = (ETH_LEN_MIN - ETH_HDR_LEN - ETH_CRC_LEN)
 
 # Ethernet payload types - http://standards.ieee.org/regauth/ethertype
+ETH_TYPE_UNKNOWN = 0x0000
 ETH_TYPE_EDP = 0x00bb  # Extreme Networks Discovery Protocol
 ETH_TYPE_PUP = 0x0200  # PUP protocol
 ETH_TYPE_IP = 0x0800  # IP protocol
@@ -51,6 +47,7 @@ ETH_TYPE_PPPoE_DISC = 0x8863  # PPP Over Ethernet Discovery Stage
 ETH_TYPE_PPPoE = 0x8864  # PPP Over Ethernet Session Stage
 ETH_TYPE_LLDP = 0x88CC  # Link Layer Discovery Protocol
 ETH_TYPE_TEB = 0x6558  # Transparent Ethernet Bridging
+ETH_TYPE_PROFINET = 0x8892  # PROFINET protocol
 
 # all QinQ types for fast checking
 _ETH_TYPES_QINQ = frozenset([ETH_TYPE_8021Q, ETH_TYPE_8021AD, ETH_TYPE_QINQ1, ETH_TYPE_QINQ2])
@@ -64,18 +61,26 @@ class Ethernet(dpkt.Packet):
 
     Attributes:
         __hdr__: Header fields of Ethernet.
-        TODO.
+            dst: (bytes): Destination MAC address
+            src: (bytes): Source MAC address
+            type: (int): Ethernet frame type (Ethernet II, Novell raw IEEE 802.3, IEEE 802.2 LLC, IEEE 802.2 SNAP)
     """
 
     __hdr__ = (
-        ('dst', '6s', ''),
-        ('src', '6s', ''),
+        ('dst', '6s', b''),
+        ('src', '6s', b''),
         ('type', 'H', ETH_TYPE_IP)
     )
     _typesw = {}
     _typesw_rev = {}  # reverse mapping
 
+    __pprint_funcs__ = {
+        'dst': mac_to_str,
+        'src': mac_to_str,
+    }
+
     def __init__(self, *args, **kwargs):
+        self._next_type = None
         dpkt.Packet.__init__(self, *args, **kwargs)
         # if data was given in kwargs, try to unpack it
         if self.data:
@@ -83,10 +88,8 @@ class Ethernet(dpkt.Packet):
                 self._unpack_data(self.data)
 
     def _unpack_data(self, buf):
-        next_type = self.type
-
         # unpack vlan tag and mpls label stacks
-        if next_type in _ETH_TYPES_QINQ:
+        if self._next_type in _ETH_TYPES_QINQ:
             self.vlan_tags = []
 
             # support up to 2 tags (double tagging aka QinQ)
@@ -94,13 +97,13 @@ class Ethernet(dpkt.Packet):
                 tag = VLANtag8021Q(buf)
                 buf = buf[tag.__hdr_len__:]
                 self.vlan_tags.append(tag)
-                next_type = tag.type
-                if next_type != ETH_TYPE_8021Q:
+                self._next_type = tag.type
+                if self._next_type != ETH_TYPE_8021Q:
                     break
             # backward compatibility, use the 1st tag
             self.vlanid, self.priority, self.cfi = self.vlan_tags[0].as_tuple()
 
-        elif next_type == ETH_TYPE_MPLS or next_type == ETH_TYPE_MPLS_MCAST:
+        elif self._next_type == ETH_TYPE_MPLS or self._next_type == ETH_TYPE_MPLS_MCAST:
             self.labels = []  # old list containing labels as tuples
             self.mpls_labels = []  # new list containing labels as instances of MPLSlabel
 
@@ -115,16 +118,20 @@ class Ethernet(dpkt.Packet):
 
             # poor man's heuristics to guessing the next type
             if compat_ord(buf[0]) == 0x45:  # IP version 4 + header len 20 bytes
-                next_type = ETH_TYPE_IP
+                self._next_type = ETH_TYPE_IP
+
+            elif compat_ord(buf[0]) & 0xf0 == 0x60:  # IP version 6 
+                self._next_type = ETH_TYPE_IP6
 
             # pseudowire Ethernet
             elif len(buf) >= self.__hdr_len__:
                 if buf[:2] == b'\x00\x00':  # looks like the control word (ECW)
                     buf = buf[4:]  # skip the ECW
-                next_type = ETH_TYPE_TEB  # re-use TEB class mapping to decode Ethernet
+                self._next_type = ETH_TYPE_TEB  # re-use TEB class mapping to decode Ethernet
 
         try:
-            self.data = self._typesw[next_type](buf)
+            eth_type = self._next_type or self.type
+            self.data = self._typesw[eth_type](buf)
             setattr(self, self.data.__class__.__name__.lower(), self.data)
         except (KeyError, dpkt.UnpackError):
             self.data = buf
@@ -133,6 +140,7 @@ class Ethernet(dpkt.Packet):
         dpkt.Packet.unpack(self, buf)
         if self.type > 1500:
             # Ethernet II
+            self._next_type = self.type
             self._unpack_data(self.data)
 
         elif (self.dst.startswith(b'\x01\x00\x0c\x00\x00') or
@@ -149,27 +157,50 @@ class Ethernet(dpkt.Packet):
             self.type = ETH_TYPE_IPX
             self.data = self.ipx = self._typesw[ETH_TYPE_IPX](self.data[2:])
 
+        elif self.type == ETH_TYPE_UNKNOWN:
+            # Unknown type, assume Ethernet
+            self._unpack_data(self.data)
+
         else:
             # IEEE 802.3 Ethernet - LLC
-            # try to unpack FCS here; we follow the same heuristic approach as Wireshark:
-            # if the upper layer len(self.data) can be fully decoded and returns its size,
-            # and there's a difference with size in the Eth header, then assume the last
-            # 4 bytes is the FCS and remaining bytes are a trailer.
+            # try to unpack FCS, padding and trailer here
+            # we follow a heuristic approach similar to that of Wireshark
+
+            # size of eth body, not including the header
             eth_len = self.len = self.type
-            if len(self.data) > eth_len:
-                tail_len = len(self.data) - eth_len
-                if tail_len >= 4:
-                    # if the last 4 bytes are zeroes that's unlikely a FCS
-                    if self.data[-4:] == b'\x00\x00\x00\x00':
-                        self.trailer = self.data[eth_len:]
-                    else:
-                        self.fcs = struct.unpack('>I', self.data[-4:])[0]
-                        self.trailer = self.data[eth_len:-4]
+
+            # actual size of the remaining data, could include eth body, padding, fcs, trailer
+            data_len = len(self.data)
+            if data_len > eth_len:
+                # everything after eth body
+                tail = self.data[eth_len:]
+
+                # could be padding + fcs, possibly trailer
+                if len(tail) > 4:
+                    # determine size of padding
+                    if eth_len < 46:  # 46=60-14; 14=size of eth hdr; all padded to 60 bytes
+                        pad_len = 46 - eth_len
+                        padding = tail[:pad_len]
+
+                        # heuristic
+                        if padding == pad_len * b'\x00':  # padding is likely zeroes
+                            self.padding = padding
+                            tail = tail[pad_len:]
+                        # else proceed to decode as fcs+trailer
+
+                # 4 bytes FCS and possible trailer
+                if len(tail) >= 4:
+                    self.fcs = struct.unpack('>I', tail[:4])[0]
+                    tail = tail[4:]
+
+                if tail:
+                    self.trailer = tail
+
             self.data = self.llc = llc.LLC(self.data[:eth_len])
 
     def pack_hdr(self):
         tags_buf = b''
-        new_type = self.type
+        new_type = self.type  # replacement self.type when packing eth header
         is_isl = False  # ISL wraps Ethernet, this determines order of packing
 
         if getattr(self, 'mpls_labels', None):
@@ -184,12 +215,18 @@ class Ethernet(dpkt.Packet):
             tags_buf = b''.join(lbl.pack_hdr() for lbl in self.mpls_labels)
 
         elif getattr(self, 'vlan_tags', None):
+            # set last tag type to next layer pointed by self.data
+            last_tag_type = self.type  # default
+            if isinstance(self.data, dpkt.Packet):
+                last_tag_type = self._typesw_rev.get(self.data.__class__, self.type)
+
             # set encapsulation types
             t1 = self.vlan_tags[0]
             if len(self.vlan_tags) == 1:
                 if isinstance(t1, VLANtag8021Q):
                     if new_type not in _ETH_TYPES_QINQ:  # preserve the type if already set
                         new_type = ETH_TYPE_8021Q
+                    t1.type = last_tag_type
                 elif isinstance(t1, VLANtagISL):
                     t1.type = 0  # 0 means Ethernet
                     is_isl = True
@@ -199,6 +236,7 @@ class Ethernet(dpkt.Packet):
                     t1.type = ETH_TYPE_8021Q
                     if new_type not in _ETH_TYPES_QINQ:
                         new_type = ETH_TYPE_8021AD
+                t2.type = last_tag_type
             else:
                 raise dpkt.PackError('maximum is 2 VLAN tags per Ethernet frame')
             tags_buf = b''.join(tag.pack_hdr() for tag in self.vlan_tags)
@@ -219,29 +257,35 @@ class Ethernet(dpkt.Packet):
         else:
             return tags_buf + hdr_buf
 
-    def __str__(self):
+    def __bytes__(self):
         tail = b''
         if isinstance(self.data, llc.LLC):
+            fcs = b''
             if hasattr(self, 'fcs'):
                 if self.fcs:
                     fcs = self.fcs
                 else:
                     # if fcs field is present but 0/None, then compute it and add to the tail
-                    fcs_buf = self.pack_hdr() + bytes(self.data) + getattr(self, 'trailer', '')
+                    fcs_buf = self.pack_hdr() + bytes(self.data)
                     # if ISL header is present, exclude it from the calculation
                     if getattr(self, 'vlan_tags', None):
                         if isinstance(self.vlan_tags[0], VLANtagISL):
                             fcs_buf = fcs_buf[VLANtagISL.__hdr_len__:]
+                    fcs_buf += getattr(self, 'padding', b'')
                     revcrc = crc32(fcs_buf) & 0xffffffff
                     fcs = struct.unpack('<I', struct.pack('>I', revcrc))[0]  # bswap32
-                tail = getattr(self, 'trailer', b'') + struct.pack('>I', fcs)
-        return str(dpkt.Packet.__bytes__(self) + tail)
+                fcs = struct.pack('>I', fcs)
+            tail = getattr(self, 'padding', b'') + fcs + getattr(self, 'trailer', b'')
+        return bytes(dpkt.Packet.__bytes__(self) + tail)
 
     def __len__(self):
         tags = getattr(self, 'mpls_labels', []) + getattr(self, 'vlan_tags', [])
         _len = dpkt.Packet.__len__(self) + sum(t.__hdr_len__ for t in tags)
-        if isinstance(self.data, llc.LLC) and hasattr(self, 'fcs'):
-            _len += len(getattr(self, 'trailer', '')) + 4
+        if isinstance(self.data, llc.LLC):
+            _len += len(getattr(self, 'padding', b''))
+            if hasattr(self, 'fcs'):
+                _len += 4
+            _len += len(getattr(self, 'trailer', b''))
         return _len
 
     @classmethod
@@ -290,24 +334,19 @@ class MPLSlabel(dpkt.Packet):
         ('_val_exp_s_ttl', 'I', 0),
     )
     # field names are according to RFC3032
+    __bit_fields__ = {
+        '_val_exp_s_ttl': (
+            ('val', 20),  # label value, 20 bits
+            ('exp', 3),   # experimental use, 3 bits
+            ('s', 1),     # bottom of stack flag, 1 bit
+            ('ttl', 8),   # time to live, 8 bits
+        )
+    }
 
     def unpack(self, buf):
         dpkt.Packet.unpack(self, buf)
-        self.val = (self._val_exp_s_ttl & 0xfffff000) >> 12  # label value, 20 bits
-        self.exp = (self._val_exp_s_ttl & 0x00000e00) >> 9   # experimental use, 3 bits
-        self.s = (self._val_exp_s_ttl & 0x00000100) >> 8     # bottom of stack flag, 1 bit
-        self.ttl = self._val_exp_s_ttl & 0x000000ff          # time to live, 8 bits
         self.data = b''
 
-    def pack_hdr(self):
-        self._val_exp_s_ttl = (
-            ((self.val & 0xfffff) << 12) |
-            ((self.exp & 7) << 9) |
-            ((self.s & 1) << 8) |
-            ((self.ttl & 0xff))
-        )
-        return dpkt.Packet.pack_hdr(self)
-
     def as_tuple(self):  # backward-compatible representation
         return (self.val, self.exp, self.ttl)
 
@@ -319,22 +358,18 @@ class VLANtag8021Q(dpkt.Packet):
         ('_pri_cfi_id', 'H', 0),
         ('type', 'H', ETH_TYPE_IP)
     )
+    __bit_fields__ = {
+        '_pri_cfi_id': (
+            ('pri', 3),  # priority, 3 bits
+            ('cfi', 1),  # canonical format indicator, 1 bit
+            ('id', 12),  # VLAN id, 12 bits
+        )
+    }
 
     def unpack(self, buf):
         dpkt.Packet.unpack(self, buf)
-        self.pri = (self._pri_cfi_id & 0xe000) >> 13   # priority, 3 bits
-        self.cfi = (self._pri_cfi_id & 0x1000) >> 12   # canonical format indicator, 1 bit
-        self.id = self._pri_cfi_id & 0x0fff           # VLAN id, 12 bits
         self.data = b''
 
-    def pack_hdr(self):
-        self._pri_cfi_id = (
-            ((self.pri & 7) << 13) |
-            ((self.cfi & 1) << 12) |
-            ((self.id & 0xfff))
-        )
-        return dpkt.Packet.pack_hdr(self)
-
     def as_tuple(self):
         return (self.id, self.pri, self.cfi)
 
@@ -345,7 +380,7 @@ class VLANtagISL(dpkt.Packet):
     __hdr__ = (
         ('da', '5s', b'\x01\x00\x0c\x00\x00'),
         ('_type_pri', 'B', 3),
-        ('sa', '6s', ''),
+        ('sa', '6s', b''),
         ('len', 'H', 0),
         ('snap', '3s', b'\xaa\xaa\x03'),
         ('hsa', '3s', b'\x00\x00\x0c'),
@@ -353,26 +388,26 @@ class VLANtagISL(dpkt.Packet):
         ('indx', 'H', 0),
         ('res', 'H', 0)
     )
+    __bit_fields__ = {
+        '_type_pri': (
+            ('type', 4),  # encapsulation type, 4 bits; 0 means Ethernet
+            ('pri', 4),   # user defined bits, 2 lo bits are used; means priority
+        ),
+        '_id_bpdu': (
+            ('id', 15),   # vlan id, 15 bits
+            ('bpdu', 1),  # bridge protocol data unit indicator
+        )
+    }
 
     def unpack(self, buf):
         dpkt.Packet.unpack(self, buf)
-        self.type = (self._type_pri & 0xf0) >> 4  # encapsulation type, 4 bits; 0 means Ethernet
-        self.pri = self._type_pri & 0x03  # user defined bits, 2 bits are used; means priority
-        self.id = self._id_bpdu >> 1  # VLAN id
-        self.bpdu = self._id_bpdu & 1
         self.data = b''
 
-    def pack_hdr(self):
-        self._type_pri = ((self.type & 0xf) << 4) | (self.pri & 0x3)
-        self._id_bpdu = ((self.id & 0x7fff) << 1) | (self.bpdu & 1)
-        return dpkt.Packet.pack_hdr(self)
-
 
 # Unit tests
 
 
 def test_eth():
-    from . import ip  # IPv6 needs this to build its protocol stack
     from . import ip6
     from . import tcp
     s = (b'\x00\xb0\xd0\xe1\x80\x72\x00\x11\x24\x8c\x11\xde\x86\xdd\x60\x00\x00\x00'
@@ -389,6 +424,18 @@ def test_eth():
     assert len(eth) == len(s)
 
 
+def test_eth_zero_ethtype():
+    s = (b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
+         b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
+         b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
+         b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x08\x89\x12\x04')
+    eth = Ethernet(s)
+    assert eth
+    assert eth.type == ETH_TYPE_UNKNOWN
+    assert str(eth) == str(s)
+    assert len(eth) == len(s)
+
+
 def test_eth_init_with_data():
     # initialize with a data string, test that it gets unpacked
     from . import arp
@@ -469,14 +516,20 @@ def test_eth_802dot1q():
 
 
 def test_eth_802dot1q_stacked():  # 2 VLAN tags
-    from . import arp
+    from binascii import unhexlify
+
+    import pytest
+
     from . import ip
-    s = (b'\x00\x1b\xd4\x1b\xa4\xd8\x00\x13\xc3\xdf\xae\x18\x81\x00\x00\x76\x81\x00\x00\x0a\x08\x00'
-         b'\x45\x00\x00\x64\x00\x0f\x00\x00\xff\x01\x92\x9b\x0a\x76\x0a\x01\x0a\x76\x0a\x02\x08\x00'
-         b'\xce\xb7\x00\x03\x00\x00\x00\x00\x00\x00\x00\x1f\xaf\x70\xab\xcd\xab\xcd\xab\xcd\xab\xcd'
-         b'\xab\xcd\xab\xcd\xab\xcd\xab\xcd\xab\xcd\xab\xcd\xab\xcd\xab\xcd\xab\xcd\xab\xcd\xab\xcd'
-         b'\xab\xcd\xab\xcd\xab\xcd\xab\xcd\xab\xcd\xab\xcd\xab\xcd\xab\xcd\xab\xcd\xab\xcd\xab\xcd'
-         b'\xab\xcd\xab\xcd\xab\xcd\xab\xcd\xab\xcd\xab\xcd')
+
+    s = unhexlify(
+        '001bd41ba4d80013c3dfae18810000768100000a0800'
+        '45000064000f0000ff01929b0a760a010a760a020800'
+        'ceb70003000000000000001faf70abcdabcdabcdabcd'
+        'abcdabcdabcdabcdabcdabcdabcdabcdabcdabcdabcd'
+        'abcdabcdabcdabcdabcdabcdabcdabcdabcdabcdabcd'
+        'abcdabcdabcdabcdabcdabcd'
+    )
     eth = Ethernet(s)
     assert eth.type == ETH_TYPE_8021Q
     assert len(eth.vlan_tags) == 2
@@ -488,12 +541,16 @@ def test_eth_802dot1q_stacked():  # 2 VLAN tags
     assert isinstance(eth.data, ip.IP)
 
     # construction
-    assert str(eth) == str(s), 'pack 1'
-    assert str(eth) == str(s), 'pack 2'
     assert len(eth) == len(s)
+    assert bytes(eth) == s
+
+    # test packing failure with too many tags
+    eth.vlan_tags += eth.vlan_tags[0]  # just duplicate the first tag
+    with pytest.raises(dpkt.PackError, match='maximum is 2 VLAN tags per Ethernet frame'):
+        bytes(eth)
 
     # construction with kwargs
-    eth2 = Ethernet(src=eth.src, dst=eth.dst, vlan_tags=eth.vlan_tags, data=eth.data)
+    eth2 = Ethernet(src=eth.src, dst=eth.dst, vlan_tags=eth.vlan_tags[:2], data=eth.data)
 
     # construction sets ip.type to 802.1ad instead of 802.1q so account for it
     assert str(eth2) == str(s[:12] + b'\x88\xa8' + s[14:])
@@ -502,6 +559,10 @@ def test_eth_802dot1q_stacked():  # 2 VLAN tags
     del eth.vlan_tags, eth.cfi, eth.vlanid, eth.priority
     assert str(eth) == str(s[:12] + b'\x08\x00' + s[22:])
 
+
+def test_eth_vlan_arp():
+    from . import arp
+
     # 2 VLAN tags + ARP
     s = (b'\xff\xff\xff\xff\xff\xff\xca\x03\x0d\xb4\x00\x1c\x81\x00\x00\x64\x81\x00\x00\xc8\x08\x06'
          b'\x00\x01\x08\x00\x06\x04\x00\x01\xca\x03\x0d\xb4\x00\x1c\xc0\xa8\x02\xc8\x00\x00\x00\x00'
@@ -527,6 +588,9 @@ def test_eth_mpls_stacked():  # Eth - MPLS - MPLS - IP - ICMP
     assert isinstance(eth.data, ip.IP)
     assert isinstance(eth.data.data, icmp.ICMP)
 
+    # exercise .pprint() for the coverage tests
+    eth.pprint()
+
     # construction
     assert str(eth) == str(s), 'pack 1'
     assert str(eth) == str(s), 'pack 2'
@@ -540,6 +604,24 @@ def test_eth_mpls_stacked():  # Eth - MPLS - MPLS - IP - ICMP
     del eth.labels, eth.mpls_labels
     assert str(eth) == str(s[:12] + b'\x08\x00' + s[22:])
 
+def test_eth_mpls_ipv6():  # Eth - MPLS - IP6 - TCP
+    from . import ip6
+    from . import tcp
+
+    s = ( b'\x00\x30\x96\xe6\xfc\x39\x00\x30\x96\x05\x28\x38\x88\x47\x00\x01'
+          b'\x01\xff\x62\x8c\xed\x7b\x00\x28\x06\xfd\x22\x22\x22\x22\x03\x3f'
+          b'\x53\xd3\x48\xfb\x8b\x5a\x41\x7f\xe6\x17\x11\x11\x11\x11\x40\x0b'
+          b'\x08\x09\x00\x00\x00\x00\x00\x00\x20\x0e\xa1\x8e\x01\xbb\xd6\xde'
+          b'\x73\x17\x00\x00\x00\x00\xa0\x02\xff\xff\x58\x7f\x00\x00\x02\x04'
+          b'\x05\x8c\x04\x02\x08\x0a\x69\x23\xe8\x63\x00\x00\x00\x00\x01\x03'
+          b'\x03\x0a\xaf\x9c\xb6\x93')
+
+    eth = Ethernet(s)
+    assert len(eth.mpls_labels) == 1
+    assert eth.mpls_labels[0].val == 16
+    assert eth.labels == [(16, 0, 255)]
+    assert isinstance(eth.data, ip6.IP6)
+    assert isinstance(eth.data.data, tcp.TCP)
 
 def test_isl_eth_llc_stp():  # ISL - 802.3 Ethernet(w/FCS) - LLC - STP
     from . import stp
@@ -554,9 +636,9 @@ def test_isl_eth_llc_stp():  # ISL - 802.3 Ethernet(w/FCS) - LLC - STP
     assert eth.vlan_tags[0].id == 333
     assert eth.vlan_tags[0].pri == 3
 
-    # check that FCS was decoded
+    # check that FCS and padding were decoded
     assert eth.fcs == 0x41c675d6
-    assert eth.trailer == b'\x00' * 8
+    assert eth.padding == b'\x00' * 8
 
     # stack
     assert isinstance(eth.data, llc.LLC)
@@ -569,11 +651,15 @@ def test_isl_eth_llc_stp():  # ISL - 802.3 Ethernet(w/FCS) - LLC - STP
 
     # construction with kwargs
     eth2 = Ethernet(src=eth.src, dst=eth.dst, vlan_tags=eth.vlan_tags, data=eth.data)
-    eth2.trailer = eth.trailer
-    eth2.fcs = None
+    eth2.padding = b'\x00' * 8
     # test FCS computation
+    eth2.fcs = None
     assert str(eth2) == str(s)
 
+    # TODO: test padding construction
+    # eth2.padding = None
+    # assert str(eth2) == str(s)
+
     # construction w/o the ISL tag
     del eth.vlan_tags, eth.vlan
     assert str(eth) == str(s[26:])
@@ -593,7 +679,7 @@ def test_eth_llc_snap_cdp():  # 802.3 Ethernet - LLC/SNAP - CDP
     # stack
     assert isinstance(eth.data, llc.LLC)
     assert isinstance(eth.data.data, cdp.CDP)
-    assert len(eth.data.data.data) == 8  # number of CDP TLVs; ensures they are decoded
+    assert len(eth.data.data.tlvs) == 8  # number of CDP TLVs; ensures they are decoded
     assert str(eth) == str(s), 'pack 1'
     assert str(eth) == str(s), 'pack 2'
     assert len(eth) == len(s)
@@ -618,7 +704,6 @@ def test_eth_llc_ipx():  # 802.3 Ethernet - LLC - IPX
 
 
 def test_eth_pppoe():   # Eth - PPPoE - IPv6 - UDP - DHCP6
-    from . import ip  # IPv6 needs this to build its protocol stack
     from . import ip6
     from . import ppp
     from . import pppoe
@@ -658,9 +743,9 @@ def test_eth_2mpls_ecw_eth_llc_stp():  # Eth - MPLS - MPLS - PW ECW - 802.3 Eth(
     eth2 = eth.data
     assert isinstance(eth2, Ethernet)
     assert eth2.len == 38  # 802.3 Ethernet
-    # no FCS, all trailer
+    # no FCS, no trailer, just 8 bytes of padding (60=38+14+8)
     assert not hasattr(eth2, 'fcs')
-    assert eth2.trailer == b'\x00' * 8
+    assert eth2.padding == b'\x00' * 8
     assert isinstance(eth2.data, llc.LLC)
     assert isinstance(eth2.data.data, stp.STP)
     assert eth2.data.data.port_id == 0x8001
@@ -702,20 +787,89 @@ def test_eth_802dot1ad_802dot1q_ip():
     assert str(e1) == str(e2)
 
 
-if __name__ == '__main__':
-    test_eth()
-    test_eth_init_with_data()
-    test_mpls_label()
-    test_802dot1q_tag()
-    test_isl_tag()
-    test_eth_802dot1q()
-    test_eth_802dot1q_stacked()
-    test_eth_mpls_stacked()
-    test_isl_eth_llc_stp()
-    test_eth_llc_snap_cdp()
-    test_eth_llc_ipx()
-    test_eth_pppoe()
-    test_eth_2mpls_ecw_eth_llc_stp()
-    test_eth_802dot1ad_802dot1q_ip()
+def test_eth_pack():
+    eth = Ethernet(data=b'12345')
+    assert str(eth)
 
-    print('Tests Successful...')
+
+def test_eth_802dot1q_with_unfamiliar_data():
+    profinet_data = (
+        b'\xfe\xff\x05\x01\x05\x01\x00\x02\x00\x00\x00\x6c\x02'
+        b'\x05\x00\x12\x00\x00\x02\x01\x02\x02\x02\x03\x02\x04\x02\x05\x02'
+        b'\x06\x01\x01\x01\x02\x02\x01\x00\x08\x00\x00\x53\x37\x2d\x33\x30'
+        b'\x30\x02\x02\x00\x22\x00\x00\x70\x6c\x63\x78\x62\x33\x30\x30\x78'
+        b'\x6b\x63\x70\x75\x78\x61\x33\x31\x37\x2d\x32\x78\x61\x70\x6e\x78'
+        b'\x72\x64\x70\x32\x32\x63\x66\x02\x03\x00\x06\x00\x00\x00\x2a\x01'
+        b'\x01\x02\x04\x00\x04\x00\x00\x02\x00\x01\x02\x00\x0e\x00\x01\xc0'
+        b'\xa8\x3c\x87\xff\xff\xff\x00\xc0\xa8\x3c\x87')
+
+    s = (b'\x00\x0c\x29\x65\x1c\x29\x00\x0e\x8c\x8a\xa2\x5e\x81\x00\x00\x00'
+         b'\x88\x92' + profinet_data)
+
+    eth = Ethernet(s)
+    assert eth.type == ETH_TYPE_8021Q
+    assert len(eth.vlan_tags) == 1
+    assert eth.vlan_tags[0].type == ETH_TYPE_PROFINET
+    assert isinstance(eth.data, bytes)
+    assert eth.data == profinet_data
+
+
+def test_eth_802dot1q_with_arp_data():  # https://github.com/kbandla/dpkt/issues/460
+    from .arp import ARP
+    e = Ethernet(src=b'foobar', dst=b'\xff' * 6)
+    v = VLANtag8021Q(pri=0, cfi=0, id=1)
+    e.vlan_tags = [v]
+    a = ARP(sha=b'foobar', spa=b'\x0a\x0a\x0a\x0a',
+            tha=b'', tpa=b'\x0a\x0a\x0a\x05')
+    e.data = a
+    assert bytes(e) == (
+        b'\xff\xff\xff\xff\xff\xfffoobar\x81\x00\x00\x01\x08\x06'  # 0x0806 = next layer is ARP
+        b'\x00\x01\x08\x00\x06\x04\x00\x01foobar\x0a\x0a\x0a\x0a'
+        b'\x00\x00\x00\x00\x00\x00\x0a\x0a\x0a\x05')
+
+
+# 802.3 Ethernet - LLC/STP - Padding - FCS - Metamako trailer
+def test_eth_8023_llc_trailer():  # https://github.com/kbandla/dpkt/issues/438
+    d = (b'\x01\x80\xc2\x00\x00\x00\x78\x0c\xf0\xb4\xd8\x91\x00\x27\x42\x42\x03\x00\x00\x02\x02\x3c'
+         b'\x00\x01\x2c\x33\x11\xf2\x39\xc1\x00\x00\x00\x02\x80\x01\x78\x0c\xf0\xb4\xd8\xbc\x80\xaa'
+         b'\x01\x00\x14\x00\x02\x00\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\xb9\x81\x20\x5c\x1e'
+         b'\x5f\xba\x3a\xa5\x47\xfa\x01\x8e\x52\x03')
+    eth = Ethernet(d)
+    assert eth.len == 39
+    assert eth.padding == b'\x00\x00\x00\x00\x00\x00\x00'
+    assert eth.fcs == 0x4db98120
+    assert eth.trailer == b'\x5c\x1e\x5f\xba\x3a\xa5\x47\xfa\x01\x8e\x52\x03'
+    assert isinstance(eth.data, llc.LLC)
+
+    # packing
+    assert bytes(eth) == d
+
+    # FCS computation
+    eth.fcs = None
+    assert bytes(eth) == d
+
+
+def test_eth_novell():
+    from binascii import unhexlify
+
+    import dpkt
+
+    buf = unhexlify(
+        '010203040506'  # dst
+        '0708090a0b0c'  # src
+        '0000'          # type (ignored)
+        'ffff'          # indicates Novell
+
+        # IPX packet
+        '0000'          # sum
+        '0001'          # len
+        '02'            # tc
+        '03'            # pt
+        '0102030405060708090a0b0c'  # dst
+        '0102030405060708090a0b0c'  # src
+    )
+
+    eth = Ethernet(buf)
+    assert isinstance(eth.data, dpkt.ipx.IPX)
+    assert eth.data.tc == 2
+    assert eth.data.data == b''
